import opusutil
import os
import glob
from stat import *

from rm_DatasetList import *
from rm_FileList import *

# constants
GROUP_MEMBER_NAME = 0
GROUP_DATA_ID = 1
ARCHIVE_CLASS = 0
ASN_CLASS = 1

# indices into file_name_format map rows
FORMAT_NAME = 0
FORMAT_REGEX = 1
FORMAT_REGEX_INDEX = 2
FORMAT_COMPILED_REGEX = 3

class FailedEXTSearch (Exception):
   pass


class rm_Mission:
  """
===========================================================================
Class: rm_Mission

Description:
------------
This class is a base class for the different kinds of missions known to
the Ingest Request Manager (rm).
It declares methods that should be overriden by the derived classes.

Members:
_name - name of mission
_archive_class_map - a map of archive class and asn class keyed on data_id
_valid_extensions - a list of valid_extension rows, each is a map of fields
_dbobj - a database object for the mission-specific database

History:
--------
01/09/04 49551 MSwam     first version
08/17/06 56017 MSwam     add fillFiles, fillDatasets
===========================================================================
  """
  def __init__(self, name):
    self._name = name
    self._archive_class_map = {}
    self._valid_extensions = None
    self._dbobj = None


  def translate_extension(self, dataset_name, ext, product_long_ext,
                          exposure_long_ext):
    #
    opusutil.PrintMsg("E","rm_Mission: translate_extension method should be "+
                          "overriden in derived class.")
  

  #-------------------------------------------------------------------------
  #
  # Name: formSearchFilespec
  #
  # Description: Forms a filespec that will be used to search for files
  #              to ingest.
  #
  # Inputs:
  #   data_id
  #
  # Outputs:
  #   format_key
  #   filespec
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def formSearchFilespec(self, rm_control, file_name_formats, data_id,
                               dataset_name, directory_path_name):
    #
    # get format name key for this data id
    format_key = rm_control[data_id]['irc_file_name_format']

    # get the file mask type using the key
    mask = file_name_formats[format_key][FORMAT_NAME]
    opusutil.PrintMsg("D","before fill,mask="+str(mask))
    #
    # fill the file mask parameters with actual values for the search
    mask = mask.replace("<DATASET_NAME>", dataset_name)
    mask = mask.replace("<EXT>", "*")
    filespec = directory_path_name + os.sep + mask
    #
    return(format_key, filespec)


  #-------------------------------------------------------------------------
  #
  # Name: prepareRegex
  #
  # Description: Prepares a regular expression and an index into the
  #              results of the expression parsing.
  #
  # Inputs:
  #   format_key
  #
  # Outputs:
  #   ext_match
  #   ext_index
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/22/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def prepareRegex(self, format_key, file_name_formats):
    #
    # get the regular expression and EXT index into the regex result
    # for parsing the EXT from the filename
    #
    ext_match = file_name_formats[format_key][FORMAT_COMPILED_REGEX]
    ext_index = int(file_name_formats[format_key][FORMAT_REGEX_INDEX])
    opusutil.PrintMsg("D","before find, regex="+
                          file_name_formats[format_key][FORMAT_REGEX]+
                          " using field "+str(ext_index))
    #
    return (ext_match, ext_index)


  #-------------------------------------------------------------------------
  #
  # Name: findFiles
  #
  # Description: Locates files for an ingesting dataset, uses a regular
  #              expression to parse the file extension from each file,
  #              translates the file extension into the extension
  #              expected by ingest, and adds the files to the dataset object.
  #
  # Inputs:
  #   datasetObj
  #   filespec
  #   ext_match
  #   ext_index
  #   product_long_ext
  #   exposure_long_ext
  #
  # Exceptions:
  #   FailedEXTSearch
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/22/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def findFiles(self, datasetObj, filespec, ext_match, ext_index,
                      product_long_ext, exposure_long_ext):
    #
    # search for files and save a file list entry for each
    opusutil.PrintMsg("D","before glob,filespec="+str(filespec))
    flist = glob.glob(filespec)

    for aFilename in flist:
      opusutil.PrintMsg("I","found "+str(aFilename))
      fname = os.path.basename(aFilename)
      opusutil.PrintMsg("D","fname="+str(fname))

      # extract the extension from the filename using the regex search
      ext = ext_match.search(fname)
      if ext == None:
        opusutil.PrintMsg("E","Failed to determine EXT for "+fname)
        raise FailedEXTSearch

      # use the index to pluck out the extension from the parsed regex
      theExtension = ext.group(ext_index)
      opusutil.PrintMsg("D","ext="+str(theExtension))

      # translate the extension for the mission, if needed
      theExtension = self.translate_extension(fname,
                                              theExtension,
                                              product_long_ext,
                                              exposure_long_ext)
      opusutil.PrintMsg("D","translated ext="+str(theExtension))

      # get file creation date as determined by 
      # mtime: Time when data was last modified.  Changed by the
      #        following functions:
      #          creat(), mknod(), pipe(), utime(), write(2)
      # 
      creation_date = os.stat(aFilename)[ST_MTIME]

      # insert a file list entry (extension MUST be uppercase)
      #
      datasetObj.addFile(
                 rm_FileList(theExtension.upper(), fname, creation_date))
    #
    return


  def get_group(self, dataset_name, data_id, request_type,
                      ingest_path, pipe_id):
    """
===========================================================================
Name: rm_Mission.get_group

Description:
------------
Looks up the group information in the mission database for a dataset
name and data_id and returns it in a list of DatasetList objects.

Inputs:
-------
dataset_name
data_id
request_type
ingest_path
pipe_id

History:
--------
01/09/04 49551 MSwam     first version
04/20/04 50823 MSwam     remove DatasetList addition for ONECLASS
06/08/04 51450 MSwam     add data_id as qualifier on get_members
===========================================================================
    """
    # get the group members
    #  (each one returned is an array of two, the DATA_ID and MEMBER_NAME)
    theGroup = self._dbobj.get_members(dataset_name, data_id)
    #
    # add them to the list, if any
    dataset_seq_number = 1
    Datasets = []
    if len(theGroup) > 0:
      opusutil.PrintMsg("D","get_group: found members...")
      for aMember in theGroup:
        opusutil.PrintMsg("D","member = "+str(aMember))

        # retrieve archive class from the map, keyed on data_id for the member
        # (each map entry is a list of two items: archive_class and asn_class)
        #
        archive_class = (
          self._archive_class_map[aMember[GROUP_DATA_ID]][ARCHIVE_CLASS])

        # create a dataset entry for the member
        d = rm_DatasetList(aMember[GROUP_MEMBER_NAME], archive_class,
                           dataset_seq_number, ingest_path, pipe_id)
        Datasets.append(d)
        dataset_seq_number += 1
      #
      # now add a single dataset entry using group name for the ASN, if needed
      #
      if request_type == "MULTICLASS":
        # for MULTICLASS retrieve the asn class from the map, keyed on data_id
        #
        archive_class = self._archive_class_map[data_id][ASN_CLASS]
        d = rm_DatasetList(dataset_name, archive_class, dataset_seq_number,
                           ingest_path, pipe_id)
        Datasets.append(d)
        dataset_seq_number += 1
    else:
      # no members found, treat like NO_ASN with a single dataset entry
      #
      opusutil.PrintMsg("I","get_group: No members for "+dataset_name+","+
                           data_id+", using dataset alone")
      # retrieve the archive class from the map, keyed on data_id
      archive_class = self._archive_class_map[data_id][ARCHIVE_CLASS]
      d = rm_DatasetList(dataset_name, archive_class, dataset_seq_number,
                         ingest_path, pipe_id)
      Datasets.append(d)
      dataset_seq_number += 1
    #
    return Datasets


  def get_total_file_count(self, Datasets):
    """
===========================================================================
Name: rm_Mission.get_total_file_count

Description:
------------
Counts up files within a dataset object.

Inputs:
-------
Datasets

Outputs:
--------
total_file_count - number of files in the datasets

History:
--------
09/13/06  56017  MSwam   first version
===========================================================================
    """
    total_file_count = 0
    for d in Datasets:
      total_file_count += d._file_count
    #
    return total_file_count


  def add_explicit_files(self, rm_control, file_name_formats,
                           data_id,  datasetObj,
                           product_long_ext, exposure_long_ext):
    """
===========================================================================
Name: rm_Mission.add_explicit_files

Description:
------------
Finds files that match the dataset_name for the passed datasetObj and
adds them to the object.  These files are "explicit" in that the datasetObj's
dataset_name is used for the file search.

Inputs:
-------
rm_control
file_name_formats
data_id
datasetObj
product_long_ext
exposure_long_ext

History:
--------
09/13/06  56017  MSwam   refactored from req_manager_main.py
===========================================================================
    """
    (format_key, filespec) = self.formSearchFilespec(
                                    rm_control,
                                    file_name_formats,
                                    data_id,
                                    datasetObj._dataset_name,
                                    datasetObj._directory_path_name)
    (ext_match, ext_index) = self.prepareRegex(format_key,
                                               file_name_formats)

    self.findFiles(datasetObj, filespec, ext_match, ext_index,
                   product_long_ext, exposure_long_ext)

    datasetObj.dump()
    return


  def add_implicit_files(self, rm_control, file_name_formats,
                           data_id,  datasetObj,
                           product_long_ext, exposure_long_ext):
    """
===========================================================================
Name: rm_Mission.add_implicit_files

Description:
------------
Finds files that match the group member names for the passed datasetObj and
adds them to the object.  These files are "implicit" in that their filenames
cannot be directly determined from the datasetObj dataset_name.  A translation
from the datasetObj dataset_name to the group member names must occur before
the file searches are performed.  In this case, a database lookup is
done to get the group member names.

Inputs:
-------
rm_control
file_name_formats
data_id
datasetObj
product_long_ext
exposure_long_ext

History:
--------
09/13/06  56017  MSwam   refactored from req_manager_main.py
===========================================================================
    """
    theGroup = self._dbobj.get_members(datasetObj._dataset_name, data_id)
    if len(theGroup) > 0:
      opusutil.PrintMsg("D","add_implicit_files: found members...")
      for aMember in theGroup:
        opusutil.PrintMsg("D","member = "+str(aMember))
        #
        (format_key, filespec) = self.formSearchFilespec(
                                          rm_control,
                                          file_name_formats,
                                          aMember[GROUP_DATA_ID],
                                          aMember[GROUP_MEMBER_NAME],
                                          datasetObj._directory_path_name)

        (ext_match, ext_index) = self.prepareRegex(format_key,
                                                   file_name_formats)

        self.findFiles(datasetObj, filespec,
                                    ext_match, ext_index,
                                    product_long_ext, exposure_long_ext)

    else:
      opusutil.PrintMsg("D","no group members found for "+str(aDataset))
    #
    return


  def fill_Files(self, rm_control, file_name_formats, dataset_name, data_id,
                           ingest_path, pipe_id,
                           product_long_ext, exposure_long_ext):
    """
===========================================================================
Name: rm_Mission.fill_Files

Description:
------------
Fills the dataset objects and then the files that correspond to each
dataset object.

Inputs:
-------
rm_control
file_name_formats
dataset_name
data_id
ingest_path
pipe_id
product_long_ext
exposure_long_ext

History:
--------
09/13/06  56017  MSwam   refactored from req_manager_main.py
===========================================================================
    """
    #
    Datasets = self.fill_Datasets(rm_control, dataset_name, data_id,
                                  ingest_path, pipe_id)
    #
    # find files for each specified dataset
    for datasetObj in Datasets:
      #
      if rm_control[data_id]['irc_request_type'] == "HIDDEN_GRP":
        self.add_implicit_files(rm_control, file_name_formats,
                                data_id,  datasetObj,
                                product_long_ext, exposure_long_ext)
      else:
        self.add_explicit_files(rm_control, file_name_formats,
                                data_id, datasetObj,
                                product_long_ext, exposure_long_ext)
    #
    # if no files were found at all, its an error
    if self.get_total_file_count(Datasets) == 0:
      opusutil.PrintMsg("E","FAILED to find any files.")
      raise NoFilesFound
    #
    return Datasets


  def fill_Datasets(self, rm_control, dataset_name, data_id,
                          ingest_path, pipe_id):
    """
===========================================================================
Name: rm_Mission.fill_Datasets

Description:
------------
Fills the dataset objects for the ingest request depending on the request
type.

Inputs:
-------
dataset_name
data_id

History:
--------
09/13/06  56017  MSwam   refactored from req_manager_main.py
===========================================================================
    """
    Datasets = []
    #
    # process the request based on request type
    dataset_seq_number = 1
    if rm_control[data_id]['irc_request_type'] == "NO_SIGNAL":
      #
      # direct use of data_id as the archive class,
      # a single dataset is specified
      d = rm_DatasetList(dataset_name, data_id.upper(), dataset_seq_number,
                         ingest_path, pipe_id)
      d.dump()
      Datasets.append(d)
      dataset_seq_number += 1

    elif (rm_control[data_id]['irc_request_type'] == "NO_ASN" or
          rm_control[data_id]['irc_request_type'] == "HIDDEN_GRP") :
      #
      # uses the mission-specific archive class map
      archive_class = self._archive_class_map[data_id][ARCHIVE_CLASS]
      #
      # a single dataset is specified
      d = rm_DatasetList(dataset_name, archive_class, dataset_seq_number,
                         ingest_path, pipe_id)
      d.dump()
      Datasets.append(d)
      dataset_seq_number += 1

    elif (rm_control[data_id]['irc_request_type'] == "ONECLASS" or
          rm_control[data_id]['irc_request_type'] == "MULTICLASS"):
      #
      Datasets = self.get_group(dataset_name, data_id,
                                      rm_control[data_id]['irc_request_type'],
                                      ingest_path, pipe_id)
      for d in Datasets:
        d.dump()

    else:
      opusutil.PrintMsg("E","fill_Datasets: Unknown request type= "+
                            str(rm_control[data_id]['irc_request_type']))
      raise UnknownRequestType
    #
    return Datasets


#========================================================================
# TEST 
# % python rm_Mission.py
#========================================================================
if __name__ == "__main__":
  a = rm_Mission("KEPLER")
  opusutil.PrintMsg("I","The next two messages SHOULD be errors...")
  a.get_group("dset","did","rtype","ipath","pipeid")
  a.translate_extension("dset","ext","plext","elext")
  opusutil.PrintMsg("I","done")
